#FIXME Depricated model currently BROKEN

import os
import json
import gzip

from .._lib import DotArray
from ._dot_array_sequence import DASequence

def load(json_file_name, zipped=False):

    if not os.path.exists(json_file_name):
        new_flname = json_file_name + ".json"
        if os.path.exists(new_flname):
            json_file_name = new_flname
        else:
            raise RuntimeError("File '{}' does not exists.".format(json_file_name))

    if zipped:
        fl = gzip.open(json_file_name, 'r')
    else:
        fl = open(json_file_name, 'r')

    rtn = DotArraySequenceArchive()
    rtn.dict = json.load(fl)
    fl.close()
    return rtn

class DotArraySequenceArchive(object):
    ## should actually use new sqlite achrive
    
    def __init__(self):
        self.dict = {}

    def add(self, dot_array):

        if isinstance(dot_array, DotArray):
            self.dict[dot_array.hash] = dot_array.as_dict()
        elif isinstance(dot_array, DASequence):
            hash_list = list(map(lambda x: x.hash, dot_array.dot_arrays))
            self.dict[dot_array.hash] = {"sequence" : hash_list}
            for da in dot_array.dot_arrays:
                self.add(da)
        else:
            RuntimeError("nsn has to be a pynsn.DotArray or a "
                         "pynsn.dot_array_sequence.DASequence")

    def remove(self, id):
        try:
            self.dict.pop(id)
        except:
            return False
        return True

    @property
    def array_ids(self):
        rtn = []
        for k,v in self.dict.items():
            if "xy" in v:
                rtn.append(k)
        return rtn

    @property
    def sequence_ids(self):
        rtn = []
        for k,v in self.dict.items():
            if "sequence" in v:
                rtn.append(k)
        return rtn

    def get_dot_array(self, id):
        try:
            d = self.dict[id]
        except:
            return None
        if "xy" not in d:
            return None # It is a DASequene

        rtn = DotArray(0,0)
        rtn.read_from_dict(d)
        return rtn

    def get_da_sequence(self, id):
        try:
            d = self.dict[id]
        except:
            return None
        if "sequence" not in d:
            return None

        tmp = []
        for id in d["sequence"]:
            tmp.append(self.get_dot_array(id))
            if tmp[-1] is None:
                raise RuntimeError("Can't find nsn {}".format(id))
        rtn = DASequence()
        rtn.append_dot_arrays(tmp)
        return rtn

    def all_properties(self):
        """eturns array with all properties and varnames

        Examples
        --------
        could be used to make a pandas dataframe

        >>> array, varnames = my_dot_array_archive.all_properties()
        >>> pandas.DataFrame(array, columns=varnames)
        """
        array = []
        feat = {}
        for id in self.array_ids:
            feat = self.get_dot_array(id)._properties.as_dict()
            array.append(list(feat.values()))

        varnames = map(lambda x:x.replace(" ", "_"), feat.keys())
        return array, list(varnames)

    def properties_csv(self, delimiter =","):

        array, varnames = self.all_properties()
        rtn = delimiter.join(varnames)
        for row in array:
            rtn += "\n" + delimiter.join(map(lambda x:str(x), row))
        return rtn

    def save(self, json_file_name, indent=None, zipped=False):

        if zipped:
            if not json_file_name.endswith(".json.gz"):
                if json_file_name.endswith(".json"):
                    json_file_name += ".gz"
                else:
                    json_file_name += ".json.gz"
            fl = gzip.open(json_file_name, "wb")

        else:
            if not json_file_name.endswith(".json"):
                json_file_name += ".json"
            fl = open(json_file_name, 'wb')

        fl.write(json.dumps(self.dict, indent=indent).encode("utf-8"))
        fl.close()




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from abc import ABCMeta, abstractmethod

import numpy as _np
from . import _colour
from .. import _lib


class ArrayDraw(metaclass=ABCMeta):
    """Generic array draw with abstract static methods

    To develop a plotter for other graphic system, inherit the abstract class
    and define you own drawing class (MyDraw())
        'get_squared_image', 'scale_image', 'draw_shape', 'draw_convex_hull'

    Image can be then generated via
    >>> MyDraw()().create_image(object_array=object_array, colours=colours)
    """

    @staticmethod
    @abstractmethod
    def get_squared_image(image_width, background_colour, **kwargs):
        """
        -------
        rtn : should return image
        """
        return

    @staticmethod
    @abstractmethod
    def scale_image(image, scaling_factor):
        pass

    @staticmethod
    @abstractmethod
    def draw_shape(image, shape, opacity, scaling_factor):
        """functions to draw object in the specific framework

        Returns
        -------
        image :  handler of plotter in the respective framework
                    (e.g. pillow image, axes (matplotlib) or svrdraw object)
        """
        pass

    @staticmethod
    @abstractmethod
    def draw_convex_hull(image, points, convex_hull_colour, opacity, scaling_factor):
        """functions to draw object in the specific framework

        Parameters
        ----------
        scaling_factor
        convex_hull_colour
        points
        image :  handler of plotter in the respective framework
                    (e.g. pillow image, axes (matplotlib) or svrdraw object)
        """
        pass


    def create_image(self, object_array, colours, antialiasing=None, **kwargs):
        """create image

        Parameters
        ----------
        object_array : the array
        colours : ImageColours
        antialiasing :   bool or number (scaling factor)
            Only useful for pixel graphics. If turn on, picture will be
            generated on a large pixel (cf. scaling factor) array and scaled
            down after generation


        Returns
        -------
        rtn : image
        """

        _lib._check_object_array(object_array)
        if colours is None:
            colours = _colour.ImageColours()
        if not isinstance(colours, _colour.ImageColours):
            raise TypeError("Colours must be of type image.ImageColours")

        if isinstance(antialiasing, bool):
            if antialiasing:  # (not if 1)
                aaf = 2  # AA default
            else:
                aaf = 1
        else:
            try:
                aaf = int(antialiasing)
            except (ValueError, TypeError):
                aaf = 1

        # prepare the pil image, make target area if required
        image_width = int(_np.ceil(object_array.target_area_radius) * 2) * aaf
        img = self.get_squared_image(image_width=image_width,
                                     background_colour=colours.background.colour,
                                     **kwargs)

        if colours.target_area.colour is not None:
            obj = _lib.Dot(xy=(0, 0), diameter=image_width,
                           attribute=colours.target_area.colour)
            self.draw_shape(img, obj, opacity=1,
                            scaling_factor=1)

        if object_array.properties.numerosity > 0:

            # draw objects
            for obj in object_array.iter_objects():
                att = obj.get_attribute_object()
                if isinstance(att, _lib.PictureFile):
                    pass
                else:  # dot or rect: force colour, set default colour if no colour
                    obj.attribute = _colour.Colour(att, colours.default_object_colour)
                self.draw_shape(img, obj, opacity=colours.opacity_object,
                                scaling_factor=aaf)

            # draw convex hulls
            if colours.field_area_positions.colour is not None and \
                    object_array.properties.field_area_positions > 0:
                self.draw_convex_hull(img,
                                      points=object_array.properties.convex_hull_positions.xy,
                                      convex_hull_colour=colours.field_area_positions,
                                      opacity=colours.opacity_guides,
                                      scaling_factor=aaf)
            if colours.field_area.colour is not None and \
                    object_array.properties.field_area > 0:
                self.draw_convex_hull(img,
                                      points=object_array.properties.convex_hull.xy,
                                      convex_hull_colour=colours.field_area,
                                      opacity=colours.opacity_guides,
                                      scaling_factor=aaf)
            #  and center of mass
            if colours.center_of_field_area.colour is not None:
                obj = _lib.Dot(xy=object_array.get_center_of_field_area(),
                               diameter=10,
                               attribute=colours.center_of_field_area.colour)
                self.draw_shape(img, obj, opacity=colours.opacity_guides,
                                scaling_factor=aaf)
            if colours.center_of_mass.colour is not None:
                obj = _lib.Dot(xy=object_array.get_center_of_mass(),
                               diameter=10,
                               attribute=colours.center_of_mass.colour)
                self.draw_shape(img, obj, opacity=colours.opacity_guides,
                                scaling_factor=aaf)

        # rescale for antialiasing
        if aaf != 1:
            img = self.scale_image(img, scaling_factor=aaf)

        return img



import numpy as np

from .._lib import geometry
from .._lib import rng


def _jitter_identical_positions(xy, jitter_size=0.1):
    """jitters points with identical position"""

    for idx, ref_object in enumerate(xy):
        identical = np.flatnonzero(np.all(np.equal(xy, ref_object), axis=1))  # find identical positions
        if len(identical) > 1:
            for x in identical:  # jitter all identical positions
                if x != idx:
                    xy[x, :] = xy[x, :] - geometry.polar2cartesian(
                        [[jitter_size, rng.generator.random() * 2 * np.pi]])[0]

    return xy


def radial_replacement_from_reference_dots(xy, ref_pos_id,
                                           neighbour_ids, replacement_size):
    """remove neighbouring position radially from reference position
    helper function, typically used for realign
    """

    # check if there is an identical position and jitter to avoid fully overlapping positions
    if np.sum(np.all(xy[neighbour_ids,] == xy[ref_pos_id, :],
                     axis=1)) > 0:
        xy = _jitter_identical_positions(xy)

    # relative polar positions to reference_dot
    tmp_polar = geometry.cartesian2polar(xy[neighbour_ids, :] - xy[ref_pos_id, :])
    tmp_polar[:, 0] = 0.000000001 + replacement_size  # determine movement size
    tmp_xy = geometry.polar2cartesian(tmp_polar)
    xy[neighbour_ids, :] = np.array([xy[neighbour_ids, 0] + tmp_xy[:, 0],
                                     xy[neighbour_ids, 1] + tmp_xy[:, 1]]).T
    return xy


def remove_overlap_from_inner_to_outer(xy, min_dist_between, distance_matrix_function):
    """returns xy and boolean, if replacements were required"""
    assert callable(distance_matrix_function)

    replacement_required = False
    # from inner to outer remove overlaps
    for i in np.argsort(geometry.cartesian2polar(xy, radii_only=True)):
        dist_mtx = distance_matrix_function(between_positions=False)
        dist = dist_mtx[i, :]
        idx_overlaps = np.flatnonzero(dist < min_dist_between).tolist()  # overlapping dot ids
        if len(idx_overlaps) > 1:
            replacement_required = True
            idx_overlaps.remove(i)  # don't move yourself
            replace_size = min_dist_between - dist[idx_overlaps]  # dist is mostly negative, because of overlap
            xy = radial_replacement_from_reference_dots(
                xy=xy,
                ref_pos_id=i,
                neighbour_ids=idx_overlaps,
                replacement_size=replace_size)
    return replacement_required


class BrownianMotion(object):

    def __init__(self, start_pos, delta=2, search_area_radius=None, bounce_boarder=True): #FIXME check delta
        """performs brownian motions (search walk) optionally in a circular area

        Parameters
        ----------

        bounce_boarder: bool
            if true, random walk bounces back at circle boarder, otherwise walk
            will be continued in the center of the area.

        Notes
        -----
        see Brownian motion https://en.wikipedia.org/wiki/Brownian_motion
        """
        if search_area_radius is not None and \
                np.hypot(start_pos[0], start_pos[1]) > search_area_radius:
            raise ValueError("start_pos is outside max_radius")

        self.area_radius = search_area_radius
        self.scale = delta ** 2
        self._history = [np.array(start_pos)]
        self.bounce = bounce_boarder

    @property
    def current(self):
        return self._history[-1]

    def step_back(self):
        """redo last step """
        if len(self._history) > 1:
            return self._history.pop()

    def center_last_step(self):
        last = self.step_back()
        if last is not None:
            new = self._history[-1] - last
            self._history.append(new)
        return self.current

    def next(self, dt=1):

        while True:  # FIXME quit criterion
            new = rng.generator.normal(loc=0, scale=self.scale * dt,
                                        size=2) + self.current
            if self.area_radius is None or \
                    np.hypot(new[0], new[1]) <= self.area_radius:
                self._history.append(new)
                return self.current

            elif not self.bounce:
                new = new - self.current # do this step from center
                self._history.append(new)
                return self.current



import json
import os

from PIL.ImageQt import ImageQt
from PyQt5.QtGui import QPixmap, QIcon
from PyQt5.QtWidgets import QAction, QMainWindow, qApp, QFileDialog

from . import dialogs
from .main_widget import MainWidget
from .sequence_display import SequenceDisplay
from .. import __version__
from .. import NSNFactory, ImageColours

from .. import distributions as distr
from .. import visual_properties as props
from ..exceptions import NoSolutionError
from ..image import _colour
from ..image import pil_image
from ..tools import _dot_array_sequence

DEFAULT_ARRAY = {"num": 40,
                 "factory": NSNFactory(target_area_radius=200, min_dist_between=2),
                 "dot_diameter": distr.Beta(mu=15, sigma=8, min_max=(5, 40)),
                 "col": ImageColours(target_area="#303030",
                                     field_area_positions=None,
                                     field_area=None,
                                     center_of_field_area=None,
                                     center_of_mass=None,
                                     default_object_colour="green",
                                     background="gray")}

ICON = {"num": 11,
        "factory": NSNFactory(target_area_radius=200),
        "dot_diameter": distr.Beta(mu=35, sigma=20, min_max=(5, 80)),
        "col": ImageColours(target_area="#3e3e3e",
                            field_area_positions=None,
                            field_area="expyriment_orange",
                            center_of_field_area=None,
                            center_of_mass=None,
                            default_object_colour="lime",
                            background=None)}


class GUIMainWindow(QMainWindow):

    def __init__(self):

        super(GUIMainWindow, self).__init__()

        self._image = None
        self.dot_array = None
        self.settings = dialogs.SettingsDialog(self, image_colours=DEFAULT_ARRAY["col"])
        self.initUI()
        self.show()

    def initUI(self):

        # menus
        exitAction = QAction(QIcon('exit.png'), '&Exit',
                             self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.setStatusTip('Exit application')
        exitAction.triggered.connect(qApp.quit)

        settingsAction = QAction('&Settings', self)
        settingsAction.triggered.connect(self.action_settings)

        saveAction = QAction('&Save current image', self)
        saveAction.setShortcut('Ctrl+S')
        saveAction.triggered.connect(self.save_array)

        printxyAction = QAction('&Print array', self)
        printxyAction.triggered.connect(self.action_print_xy)
        printparaAction = QAction('&Print parameter', self)
        printparaAction.triggered.connect(self.action_print_para)

        adaptAction = QAction('&Adapt property', self)
        adaptAction.triggered.connect(self.action_adapt)

        sequenceAction = QAction('&Make _sequence', self)
        sequenceAction.triggered.connect(self.action_make_sequence)

        aboutAction = QAction('&About', self)

        # self.statusBar()
        menubar = self.menuBar()

        arrayMenu = menubar.addMenu('&Array')
        arrayMenu.addAction(settingsAction)
        arrayMenu.addAction(saveAction)
        arrayMenu.addSeparator()
        arrayMenu.addAction(exitAction)

        toolMenu = menubar.addMenu('&Tools')
        toolMenu.addAction(sequenceAction)
        toolMenu.addAction(adaptAction)
        toolMenu.addSeparator()
        toolMenu.addAction(printparaAction)
        toolMenu.addAction(printxyAction)

        aboutMenu = menubar.addMenu('&About')
        aboutMenu.addAction(aboutAction)

        # main widget
        factory = DEFAULT_ARRAY["factory"]
        factory.set_appearance_dot(diameter=DEFAULT_ARRAY["dot_diameter"])
        self.main_widget = MainWidget(self, self.settings, DEFAULT_ARRAY["num"],
                                      factory)
        self.setCentralWidget(self.main_widget)
        self.main_widget.btn_generate.clicked.connect(self.action_generate_btn)
        self.main_widget.dot_colour.edit.editingFinished.connect(self.action_dot_colour_change)

        self.move(300, -300)
        self.setWindowTitle('PyNSN GUI {}'.format(__version__))

        # ICON
        factory = ICON["factory"]
        factory.set_appearance_dot(diameter=ICON["dot_diameter"])
        oa = factory.create_random_array(n_objects=ICON["num"])
        self._image = pil_image.create(object_array=oa,
                                       colours=ICON["col"], antialiasing=True)

        self.setWindowIcon(QIcon(self.pixmap()))
        self._image = None
        self.action_generate_btn()

    def make_new_array(self):
        factory = self.get_factory()
        try:
            self.dot_array = factory.create_random_array(
                n_objects=self.get_number())
        except NoSolutionError as error:
            self.main_widget.text_error_feedback(error)
            raise error

        if self.settings.bicoloured.isChecked():
            data_array2 = random_array.create(n_objects=self.main_widget.number2.value,
                                              array_parameter=ref_ar,
                                              size_distribution=sdr,
                                              occupied_space=self.dot_array)
            data_array2.set_attributes(self.main_widget.dot_colour2.text)
            self.dot_array.join(data_array2)

        self.dot_array.mod_round_values(decimals=self.settings.rounding_decimals.value)
        self._image = None

    def image(self):
        if self._image is not None:
            return self._image
        else:
            para = self.get_image_colours()
            image_colours = _colour.ImageColours(
                target_area=para.target_area,
                field_area_positions=para.field_area_positions,
                field_area=para.field_area,
                center_of_field_area=para.center_of_field_area,
                center_of_mass=para.center_of_mass,
                default_object_colour=para.default_object_colour,
                background=para.background)

            self._image = pil_image.create(object_array=self.dot_array,
                                           colours=image_colours,
                                           antialiasing=self.settings.antialiasing.isChecked())
            # todo maybe: gabor_filter=ImageFilter.GaussianBlur

            return self._image

    def get_number(self):
        return self.main_widget.number.value

    def get_factory(self):
        rtn = NSNFactory(
            target_area_radius=self.main_widget.target_area_radius.value,
            min_dist_between=self.main_widget.min_dist_between.value)
        rtn.set_appearance_dot(
            diameter=distr.Beta(mu=self.main_widget.item_diameter_mean.value,
                                sigma=self.main_widget.item_diameter_std.value,
                                min_max=[self.main_widget.item_diameter_range.value1,
                                         self.main_widget.item_diameter_range.value2]))

        return rtn

    def get_image_colours(self):
        # check colour input
        try:
            colour_area = _colour.Colour(self.settings.colour_area.text)
        except TypeError:
            colour_area = None
            self.settings.colour_area.text = "None"
        try:
            colour_convex_hull_positions = _colour.Colour(
                self.settings.colour_convex_hull_positions.text)
        except TypeError:
            colour_convex_hull_positions = None
            self.settings.colour_convex_hull_positions.text = "None"
        try:
            colour_convex_hull_dots = _colour.Colour(
                self.settings.colour_convex_hull_dots.text)
        except TypeError:
            colour_convex_hull_dots = None
            self.settings.colour_convex_hull_dots.text = "None"
        try:
            colour_background = _colour.Colour(
                self.settings.colour_background.text)
        except TypeError:
            colour_background = None
            self.settings.colour_background.text = "None"

        return _colour.ImageColours(target_area=colour_area,
                                    field_area_positions=colour_convex_hull_positions,
                                    field_area=colour_convex_hull_dots,
                                    center_of_field_area=None,
                                    center_of_mass=None,
                                    default_object_colour=self.settings.default_object_colour,
                                    background=colour_background)

    def pixmap(self):
        return QPixmap.fromImage(ImageQt(self.image()))

    def show_current_image(self, remake_image=False):
        """"""
        if remake_image:
            self._image = None
        factory = self.get_factory()
        w = factory.target_area_radius * 2
        self.main_widget.resize_fields(width=w, text_height=150)
        self.main_widget.picture_field.setPixmap(self.pixmap())
        self.main_widget.adjustSize()
        self.adjustSize()

    def action_generate_btn(self):
        """"""
        self.make_new_array()
        self.show_current_image(remake_image=True)
        self.write_properties()
        self.main_widget.updateUI()

    def write_properties(self, clear_field=True):
        txt = self.dot_array._properties.as_text(extended_format=True,
                                                 with_hash=True)
        if self.settings.bicoloured.isChecked():
            for da in self.dot_array.get_split_arrays():
                txt += "Attribute {}\n".format(da._attributes[0])
                txt += da._properties.as_text(extended_format=True,
                                              with_hash=False)
        if clear_field:
            self.main_widget.text_clear()

        self.main_widget.text_out(txt)

    def action_print_xy(self):
        """"""
        txt = self.dot_array.csv(hash_column=False, num_idx_column=False, colour_column=True)
        self.main_widget.text_out(txt)

    def action_print_para(self):
        d = {'number': self.get_number()}
        d['image_parameter'] = self.get_image_colours().as_dict()
        ref_array, size_distr = self.get_factory()
        d['reference_array'] = ref_array.as_dict()
        d['size_distribution'] = size_distr.as_dict()
        self.main_widget.text_out("# parameter\n" + json.dumps(d, indent=2))

    def save_array(self):
        """"""

        filename, extension = QFileDialog.getSaveFileName(
            self,
            caption='Save file',
            directory=".",
            filter="Image PNG File (*.png);; Image BMP File (*.bmp);; JSON File (.json)")

        if len(filename) > 0:
            filename = os.path.abspath(filename)

            ext = extension.split("(")[1].replace(")", "") \
                .replace("*", "").strip()
            if not filename.endswith(ext):
                filename = filename + ext

            if ext == ".json":
                self.dot_array.save(filename)

            elif ext == ".png" or ext == ".bmp":
                self.image().save(filename)

    def action_adapt(self):
        """"""
        prop = self.dot_array._properties.as_dict()
        prop, value = dialogs.AdaptPropertyDialog.get_response(self, prop)
        if prop is not None:
            self.dot_array = props.fit.visual_property(self.dot_array,
                                                       prop, value=value)
            if prop in [x for x in props.flags if x.is_size_property()]:
                self.dot_array.center_array()
                self.dot_array.realign()
            self.show_current_image(remake_image=True)
            self.write_properties()
            self.main_widget.updateUI()

    def action_settings(self):
        """"""
        result = self.settings.exec_()
        self.main_widget.updateUI()
        self.action_generate_btn()
        # self.show_current_image(remake_image=True)

    def action_dot_colour_change(self):
        """
        """
        try:
            colour_dot = _colour.Colour(self.main_widget.dot_colour.text)
            self.settings.default_object_colour = colour_dot
        except TypeError:
            colour_dot = self.settings.default_object_colour
            self.main_widget.dot_colour.text = colour_dot.colour

        self.dot_array.set_attributes(colour_dot.colour)
        self.show_current_image(remake_image=True)

    def action_make_sequence(self):
        adapt_methods, adapt_range, extra_space = \
            dialogs.SequenceDialog.get_response(self)
        adapt_methods = adapt_methods[0]  # FIXME just adapt the first

        d = {"adapt range": adapt_range,
             "extra_space": extra_space}
        d["adapt_methods"] = adapt_methods
        self.main_widget.text_out("# Sequence\n" + \
                                  json.dumps(d))
        ref_array, sdr = self.get_factory()
        ref_array.min_dist_area_boarder = extra_space / 2
        ref_array.target_area_radius += ref_array.min_dist_area_boarder

        if adapt_methods is not None:
            sequence = _dot_array_sequence.create(
                specs=specs,
                adapt_property=adapt_methods,
                adapt_value=self.dot_array._properties.get(adapt_methods),
                min_max_numerosity=adapt_range)
            SequenceDisplay(self, da_sequence=sequence,
                            start_numerosity=self.dot_array._properties.numerosity,
                            image_colours=self.get_image_colours(),
                            antialiasing=self.settings.antialiasing.isChecked()).exec_()




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'


import math as _math
from .picture_file import PictureFile
from ..image._colour import Colour


class Point(object):

    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def __repr__(self):
        return "Point(xy={})".format(self.xy)

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        return Point(self.x * other, self.y * other)

    def __div__(self, other):
        return Point(self.x / other if other else self.x,
                            self.y / other if other else self.y)

    def __iadd__(self, other):
        self.x += other.x
        self.y += other.y
        return self

    def __isub__(self, other):
        self.x -= other.x
        self.y -= other.y
        return self

    def __imul__(self, other):
        self.x *= other
        self.y *= other
        return self

    def __idiv__(self, other):
        self.x /= other
        self.y /= other
        return self

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y

    @property
    def xy(self):
        return self.x, self.y

    @xy.setter
    def xy(self, xy_tuple):
        self.x = xy_tuple[0]
        self.y = xy_tuple[1]

    @property
    def polar_radius(self):
        return _math.hypot(self.x, self.y)

    @polar_radius.setter
    def polar_radius(self, value):
        self.polar = (value, self.polar_angle)

    @property
    def polar_angle(self):
        return _math.atan2(self.y, self.x)

    @polar_angle.setter
    def polar_angle(self, value):
        self.polar = (self.polar_radius, value)

    @property
    def polar(self):
        """polar coordinate (radius, pos_angle) """
        return self.polar_radius, self.polar_angle

    @polar.setter
    def polar(self, rad_ang):
        """polar coordinate (radius, angle) """

        self.x = rad_ang[0] * _math.cos(rad_ang[1])
        self.y = rad_ang[0] * _math.sin(rad_ang[1])

    def distance(self, d):
        """Returns Euclidean distance to the another Coordinate. The function
        does not takes the size of an object into account.

        Parameters
        ----------
        d : Point

        Returns
        -------
        distance : float

        """

        return _math.hypot(self.x - d.x, self.y - d.y)


class ShapeAttribute(object):

    def __init__(self, attribute):
        self._attribute = None
        self.attribute = attribute # setter

    @property
    def attribute(self):
        return self._attribute

    @attribute.setter
    def attribute(self, attr):
        """set attribute 
        
        Parameters
        ----------
        attr : anything
            can be, in principle, anything.
            If Colour or PictureFile, it will convert it to their string
            representation
        """
        if isinstance(attr, Colour):
            self._attribute = attr.colour
        elif isinstance(attr, PictureFile):
            self._attribute = attr.attribute
        else:
            self._attribute = attr

    def get_attribute_object(self):
        """Class instance of the attribute, if possible

        Returns
        -------
        rtn : attribute
            If attribute represents Colour or PictureFile it returns the instance
            of the respective class otherwise the previously defined  attribute
        """

        if isinstance(self._attribute, str):
            # check is color or picture
            col = Colour(self._attribute)
            if col.colour is not None:
                return col
            else:
                fl_name = PictureFile.check_attribute(self._attribute)
                if fl_name is not None:
                    return PictureFile(fl_name)
        return self._attribute


class Dot(Point, ShapeAttribute):

    def __init__(self, xy, diameter, attribute=None):
        """Initialize a point

        Handles polar and cartesian representation (optimised processing, i.e.,
        conversions between coordinates systems will be done only once if needed)

        Parameters
        ----------
        xy : tuple of two numeric
        diameter : numeric
        attribute : attribute (string, optional)
        """
        if isinstance(attribute, PictureFile):
            raise NotImplementedError("Dot arrays can not handle pictures.")

        Point.__init__(self, x=xy[0], y=xy[1])
        ShapeAttribute.__init__(self, attribute)
        self.diameter = diameter

    def __repr__(self):
        return "Dot(xy={}, diameter={}, attribute='{}')".format(self.xy,
                            self.diameter, self.attribute)

    def distance(self, d):
        """Return Euclidean distance to the dot d. The function takes the
        diameter of the points into account.

        Parameters
        ----------
        d : Dot

        Returns
        -------
        distance : float

        """

        return Point.distance(self, d) - \
               ((self.diameter + d.diameter) / 2.0)

    @property
    def area(self):
        return _math.pi * (self.diameter ** 2) / 4.0

    @property
    def perimeter(self):
        return _math.pi * self.diameter


class Rectangle(Point, ShapeAttribute):

    def __init__(self, xy, size, attribute=None):
        """Initialize a Rectangle

        Handles polar and cartesian representation (optimised processing, i.e.,
        conversions between coordinates systems will be done only once if needed)

        Rectangle can also consist of a picture

        Parameters
        ----------
        xy : tuple
            tuple of two numeric (default=(0, 0))
        size : tuple
            tuple of two numeric (default=(0, 0))
        attribute : attribute (string) or PictureFile
        """

        Point.__init__(self, x=xy[0], y=xy[1])
        ShapeAttribute.__init__(self, attribute)
        self.width, self.height = size

    def __repr__(self):
        return "Rectangle(xy={}, size={}, attribute='{}')".format(self.xy,
                                    self.size, self.get_attribute_str())

    @property
    def left(self):
        return self.x - 0.5 * self.width

    @property
    def top(self):
        return self.y + 0.5 * self.height

    @property
    def right(self):
        return self.x + 0.5 * self.width

    @property
    def bottom(self):
        return self.y - 0.5 * self.height

    def iter_edges(self, xy_tuple=False):
        """iterator over all four edges

        Returns
        -------
        iterator over Points or tuple (x,y)
        """
        for x in [Point(self.left, self.top),
                  Point(self.right, self.top),
                  Point(self.right, self.bottom),
                  Point(self.left, self.bottom)]:
            if xy_tuple:
                yield x.xy
            else:
                yield x

    @property
    def size(self):
        return self.width, self.height

    @size.setter
    def size(self, size):
        self.width, self.height = size

    def is_point_inside_rect(self, xy):
        return (self.left <= xy[0] <= self.right and
                self.top <= xy[1] <= self.bottom)

    def overlaps_with(self, rect):
        d = self.xy_distances(rect)
        return not(d[0]>0 or d[1]>0)

    @property
    def proportion(self):
        """Proportion of the rectangle (width/height)"""
        return self.width/self.height

    @property
    def area(self):
        return self.width * self.height

    @property
    def perimeter(self):
        return 2 * (self.width + self.height)

    @property
    def diagonal(self):
        return _math.sqrt(self.width**2 + self.height**2)

    def xy_distances(self, other):
        """return distances on both axes between rectangles. 0 indicates
        overlap off edges along that dimension.
        """
        assert isinstance(other, Rectangle)
        #overlaps in x or y
        pos_dist = abs(self.x - other.x), abs(self.y - other.y)
        max_overlap_dist = (self.width + other.width)/2, (self.height + other.height)/2
        if  pos_dist[0] <= max_overlap_dist[0]:
            dx = 0
        else:
            dx = pos_dist[0] - max_overlap_dist[0]

        if  pos_dist[1] <= max_overlap_dist[1]:
            dy = 0
        else:
            dy = pos_dist[1] - max_overlap_dist[1]

        return dx, dy

    def distance(self, other):
        """Return Euclidean distance to the dot d. The function takes the
        size of the rectangles into account.

        Parameters
        ----------
        other : Rectangle

        Returns
        -------
        distance : float

        """
        dx, dy = self.xy_distances(other=other)
        return _math.hypot(dx, dy)




#!/usr/bin/env python3
"""
Installer
"""

from setuptools import setup, find_packages
import codecs
import os
from sys import version_info as _vi

package_name = "pynsn"

if _vi.major < 3 or _vi.minor < 6:
    raise RuntimeError("{0} requires Python 3.6 or larger.".format(package_name))

install_requires = ["numpy>=1.19",
                    "scipy>=1.5",
                    "Pillow>=8.4",
                    "svgwrite>=1.4"]

extras_require = {
    'gui':                ["PyQt5>=5.14"],
    'pygame':             ["pygame>=1.9"],
    'expyriment':         ["expyriment>=0.9"],
    'matplotlib':         ["matplotlib>=3.2"]
}

entry_points = {'console_scripts': ['pynsn=pynsn.gui:start']}

packages = find_packages(".")


def readme():
    directory = os.path.dirname(os.path.join(
        os.getcwd(), __file__, ))
    with codecs.open(
        os.path.join(directory, "README.md"),
        encoding="utf8",
        mode="r",
        errors="replace",
        ) as file:
        return file.read()


def get_version(package):
    """Get version number"""

    with open(os.path.join(package, "__init__.py")) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split("'")[1]
    return "None"


if __name__ == '__main__':
    setup(
        name = package_name,
        version=get_version(package_name),
        description='Creating Non-Symbolic Number Displays',
        author='Oliver Lindemann',
        author_email='lindemann@cognitive-psychology.eu',
        license='GNU GPLv3',
        url='https://github.com/lindemann09/PyNSN',
        packages=packages,
        include_package_data=True,
        setup_requires=[],
        install_requires=install_requires,
        entry_points=entry_points,
        extras_require=extras_require,
        keywords = "", #ToDo
        classifiers=[
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering"
        ],
        long_description=readme(),
        long_description_content_type='text/markdown'
    )



from .base_classes import ArrayParameter, AttributeArray
from .dot_array import DotArray
from .rect_array import RectangleArray
from .shapes import Point, Dot, Rectangle
from .picture_file import PictureFile
from .appearance_sampler import AppearanceSampler
from .factory import NSNFactory


# helper for type checking and error raising error
def _check_array_parameter(obj):
    if not isinstance(obj, ArrayParameter):
        raise TypeError("DotArray, RectangleArray or ArrayParameter expected, but not {}".format(
            type(obj).__name__))


def _check_object_array(obj):
    if not isinstance(obj, (DotArray, RectangleArray)):
        raise TypeError("DotArray or RectangleArray expected, but not {}".format(
            type(obj).__name__))

